--[[
	Todo
		- confirm GetThreat() works with aloft.
		
	Notes
		- hooking plates OnEnter requires the plates enable mouse. But doing that stops the mouseover ID being available. 
		I've tried disabing mouse once OnEnter fires but it's not guaranteed that the UPDATE_MOUSEOVER will then fire with the mobs moueoverID. This is more true when nameplates are clusterd ontop of each other.
		- TidyPlates sometimes removes the children from a plate.
		- TidyPlates seems to break my OnShow hook.
]]

local MAJOR, MINOR = "LibNameplate-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") return end

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
if not lib.callbacks then	error(MAJOR .. " CallbackHandler-1.0.") return end

-- Globals
local _G					= _G
local WorldFrame_			= WorldFrame
local select_				= select
local UnitExists_			= UnitExists
local UnitGUID_				= UnitGUID
local UnitName_				= UnitName
local string_gsub			= string.gsub
local pairs_				= pairs
local table_remove			= table.remove
local table_insert			= table.insert
local GetMouseFocus_		= GetMouseFocus
local GetNumRaidMembers_	= GetNumRaidMembers
local GetNumPartyMembers_	= GetNumPartyMembers
local UnitIsUnit_			= UnitIsUnit
local GetRaidTargetIndex_	= GetRaidTargetIndex
local UnitInParty_			= UnitInParty
local tonumber_				= tonumber
local print_				= print
local math_floor			= math.floor
local RAID_CLASS_COLORS_	= RAID_CLASS_COLORS
local table_getn			= table.getn
local CreateFrame_			= CreateFrame
local _ --yes a underscore, else it'll go global. =/

--Conts
local scanDelay		= 1 -- Scan for new nameplates every 1 seconds. We hook OnShow so this doesn't need to be done rapidly.
local updateDelay	= 1 -- Update nameplate info every 1 seconds. This checks if a nameplate has a raid icon then trys to find the GUID to that icon.

local aloftRunning			= false
local tidyPlatesRunning 	= false


local DEBUG = false
--[===[@debug@
DEBUG = true
--@end-debug@]===]

local function CmdHandler()
	DEBUG = not DEBUG
end
_G.SlashCmdList["LIBNAMEPLATEDEBUG"] = CmdHandler
_G.SLASH_LIBNAMEPLATEDEBUG1 = "/lnbug"
local function debugPrint(...)
	if DEBUG == true then
		print_(...)
	end
end

local onShowHooks = {}
local onHideHooks = {}
local onUpdateHooks = {}

local regionNames = { "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon" }
local regions = {}
for i, name in pairs_(regionNames) do 
	regions[name] = i
end

local barNames = { "healthBar", "castBar"}
local bars = {}
for i, name in pairs_(barNames) do 
	bars[name] = i
end



--"Star", "Circle", "Diamond", "Triangle", "Moon", "Square", "Cross", "Skull"
--~ local raidIconTexture = "Interface\\TargetingFrame\\UI-RaidTargetingIcons"
--			= left							right
--			= ULx	ULy		LLx		LLy		URx		URy		LRx		LRy
-- star		= 0		0		0		0.25	0.25	0		0.25	0.25
-- circle	= 0.25	0		0.25	0.25	0.5		0		0.5		0.25
-- diamond	= 0.5	0		0.5		0.25	0.75	0		0.75	0.25
-- triangle	= 0.75	0		0.75	0.25	1		0		1		0.25
-- moon		= 0		0.25	0		0.5		0.25	0.25	0.25	0.5
-- square 	= 0.25	0.25	0.25	0.5		0.5		0.25	0.5		0.5
-- cross	= 0.5	0.25	0.5		0.5		0.75	0.25	0.75	0.5
-- skull	= 0.75	0.25	0.75	0.5		1		0.25	1		0.5
local raidIconTexCoord = {--from GetTexCoord. input is ULx and ULy (first 2 values).
	[0]		= {
		[0]		= 1, --star 
		[0.25]	= 5, --moon
	},
	[0.25]	= {
		[0]		= 2, --circle 
		[0.25]	= 6, --square
	},
	[0.5]	= {
		[0]		= 3, --star 
		[0.25]	= 7, --cross
	},
	[0.75]	= {
		[0]		= 4, --star 
		[0.25]	= 8, --skull
	},
}

lib.nameplates		= lib.nameplates or {}
lib.GUIDs			= lib.GUIDs or {}
lib.names			= lib.names or {}

local enableMouseFrames = {}

local plateGUIDs = {}

local colorToClass = {}
local function pctToInt(number) return math_floor((100*number) + 0.5) end
for classname, color in pairs_(RAID_CLASS_COLORS_) do
	colorToClass["C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b)] = classname
--~ 	debugPrint("Class", classname, "C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b))
end


--[[
	these always return nil
	frame:IsClampedToScreen(), frame:IsIgnoringDepth(), frame:IsMovable(), frame:IsToplevel(), frame:GetDepth(), frame:GetEffectiveDepth(), 

]]
local function IsNamePlateFrame(frame)
	if frame:GetName() then
		debugPrint("IsNamePlateFrame","No name!")
		return false
	end

	if frame:GetID() ~= 0 then
		return false
	end
	
	if frame:GetObjectType() ~= "Frame" then
		debugPrint("IsNamePlateFrame","Not a frame!")
		return false
	end
		
	if frame:GetNumChildren() == 0 then
		
		debugPrint("IsNamePlateFrame","No children?!!", frame:GetName(), frame.extended and frame.extended.bars)
		return false
	end
		
	local children = {frame:GetChildren()}
	local sbFound = false
	for c = 1, table_getn(children) do
		if children[c]:GetObjectType() == "StatusBar" then
			sbFound = true
			break
		end
	end
	if sbFound == false then
		debugPrint("IsNamePlateFrame","No statusbar!")
		return false
	end

	return true
end



--[===[@debug@
function LC_TEST()--		/run LC_TEST()
	local c = {WorldFrame_:GetChildren()}
	local frame, cFrame
	for i = 1, table_getn(c) do
		frame = c[i]
		if frame:IsShown() then
			--frame.extended
			debugPrint(IsNamePlateFrame(frame), "ID:"..tostring(frame:GetID()), "C:"..tostring(frame:GetNumChildren()), "R:"..tostring(frame:GetNumRegions()))
			
		end
	end
end
--@end-debug@]===]

local isOnScreen = {}

local function FindNameplates()
--~ 	debugPrint("FindNameplates ")
	local frames = {WorldFrame_:GetChildren()}
	local frame
	for i = 1, table_getn(frames) do
		frame = frames[i]
		if frame:IsShown() then
			if IsNamePlateFrame(frame) then
				if not lib.nameplates[frame] then
					lib:SetupNameplate(frame, true)
				elseif not isOnScreen[frame] then
					debugPrint("The hell? Did our OnShow hook fail?")
					onShowHooks[frame] = false
					lib:SetupNameplate(frame, true)
				end
			end
		end
	end
--~ 	debugPrint("FindNameplates")
end

local function threatByColor( object , name)
	local redCan, greenCan, blueCan, alphaCan = object:GetVertexColor()
	local shown
	if not object:IsShown() then return "LOW" end
	if greenCan > .7 then return "MEDIUM" end
	if redCan > .7 then return "HIGH" end
end

local function reactionByColor(r, g, b, a)																											
	if r < .01 and b < .01 and g > .99 then return "FRIENDLY", "NPC" 
	elseif r < .01 and b > .99 and g < .01 then return "FRIENDLY", "PLAYER"
	elseif r > .99 and b < .01 and g > .99 then return "NEUTRAL", "NPC"
	else return "HOSTILE", "PLAYER" end
end

local function combatByColor(r, g, b, a) 
	return (r > .5 and g < .5)
end

function lib:GetName(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.name
	end

	--string_gsub returns 2 values. I just want to return 1.
	local name = string_gsub(select_(regions.nameText, frame:GetRegions()):GetText(), "|cffffffff", "")
	return name
end

function lib:GetLevel(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.level
	end
	return select_(regions.levelText, frame:GetRegions()):GetText()
end

function lib:GetReaction(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.reaction
	elseif aloftRunning == true and frame.aloftData then
		local r, g, b = frame.aloftData.originalHealthBarR, frame.aloftData.originalHealthBarG, frame.aloftData.originalHealthBarB
		return reactionByColor(r, g, b, 1)
	end
	return reactionByColor(select_(bars.healthBar, frame:GetChildren()):GetStatusBarColor())
end

function lib:GetType(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.type
	end
	return select_(2, reactionByColor(select_(bars.healthBar, frame:GetChildren()):GetStatusBarColor()))
end

function lib:IsBoss(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isBoss
	end
	return select_(regions.dangerSkull, frame:GetRegions()):IsShown()
end

function lib:GetHealthMax(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.healthmax
	end
	local _, max = select_(bars.healthBar, frame:GetChildren()):GetMinMaxValues()
	return max
end

--This will return nil if we're not in a PvP zone (like in citys)
function lib:GetClass(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.class
	end
	local red, green, blue = select_(bars.healthBar, frame:GetChildren()):GetStatusBarColor()
	return colorToClass["C"..pctToInt(red)+pctToInt(green)+pctToInt(blue)]
end

function lib:IsElite(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isElite
	end
	return select_(regions.eliteIcon, frame:GetRegions()):IsShown()
end

function lib:GetThreatSituation(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.threatSituation
	end
	return threatByColor(select_(regions.threatGlow, frame:GetRegions()))
end

function lib:GetAlpha(frame)-- really?
	return frame:GetAlpha()
end

function lib:IsTarget(frame)
	return frame:IsShown() and frame:GetAlpha() == 1 and UnitExists_("target")
end

function lib:GetHealth(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.health
	end
	return select_(bars.healthBar, frame:GetChildren()):GetValue()
end

function lib:IsMouseover(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isMouseover
	end
	return select_(regions.highlightTexture, frame:GetRegions()):IsShown()
end

function lib:IsCasting(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isCasting
	end
	return select_(bars.castBar, frame:GetChildren()):IsShown()
end

function lib:IsInCombat(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isInCombat
	end
	return combatByColor(select_(regions.nameText, frame:GetRegions()):GetTextColor())
end

function lib:IsMarked(frame)
	if tidyPlatesRunning == true and frame.extended then
		return frame.extended.unit.isMarked
	end
	return select_(regions.raidIcon, frame:GetRegions()):IsShown()
end

function lib:GetRaidIcon(frame)
	local raidIcon
	if tidyPlatesRunning == true and frame.extended then
		raidIcon = frame.extended.regions.raidIcon
	else
		raidIcon = select_(regions.raidIcon, frame:GetRegions())
	end
	if raidIcon:IsShown() then
		local ULx,ULy = raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
		if ULx and ULy then
			return raidIconTexCoord[ULx] and raidIconTexCoord[ULx][ULy] or 0
		end
	end
	return 0
end

function lib:GetGUID(frame)
	return plateGUIDs[frame]
end

-- ***
function lib:GetTargetNameplate()
	if self.targeted and self.targeted:IsShown() then
		return self.targeted
	end
end

function lib:GetNameplateByGUID(GUID)
	if self.GUIDs[GUID] and self.GUIDs[GUID]:IsShown() then
		return self.GUIDs[GUID]
	end
end


function lib:GetNameplateByName(name)
	if self.names[name] and self.names[name]:IsShown() then
		return self.names[name]
	end
end

function lib:GetNameplateByUnit(unitID)
	if UnitIsUnit_(unitID, "target") then
		return self:GetTargetNameplate()
	end
	local GUID = UnitGUID_(unitID)
	if self.GUIDs[GUID] then
		return self.GUIDs[GUID], self.GUIDs[GUID].LibNP
	end
	local name = UnitName_(unitID)
	if self.names[GUID] then
		return self.names[GUID], self.names[GUID].LibNP
	end
end

local function RecycleNameplate(frame)
	lib.callbacks:Fire("LibNameplate_RecycleNameplate", frame)
	local LibNP = lib.nameplates[frame]
	if LibNP then
		local GUID = plateGUIDs[frame]
		local name = LibNP.name
		
		if lib.GUIDs[GUID] then lib.GUIDs[GUID] = nil end
		if lib.names[name] then lib.names[name] = nil end
	end

	plateGUIDs[frame] = false
	
--~ 	debugPrint("RecycleNameplate", onShowHooks[frame], onShowHooks[frame]:GetParent(), onShowHooks[frame]:GetParent() == frame)
--~ 	onShowHooks[frame]:Show()
end

local function FoundPlateGUID(frame, GUID, unitID)
	plateGUIDs[frame] = GUID
	
	lib.GUIDs[GUID] = frame
	lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID)
end

local function GetMouseoverGUID(frame)
	local unitID = "mouseover"
	if UnitExists_(unitID) then
		FoundPlateGUID(frame, UnitGUID_(unitID), unitID)
	end
end

local function FindPlateWithRaidIcon(iconNum, unitID)
	local GUID = UnitGUID_(unitID)
	local LibNP
	for frame, LibNP in pairs_(lib.nameplates) do 
		if lib:IsMarked(frame) and lib:GetRaidIcon(frame) == iconNum then
			return frame, LibNP
		end
	end
	return nil
end



local function FindGUIDByRaidIcon(frame, raidNum)
	local LibNP = lib.nameplates[frame]
	local group, num = "", 0
	if GetNumRaidMembers_() > 1 then
		group, num = "raid", GetNumRaidMembers_()
	elseif GetNumPartyMembers_() > 0 then
		group, num = "party", GetNumPartyMembers_()
	end
	
	local unitID, targetID, targetIndex
	for i = 1, num do
		unitID = "party"..i;
		targetID = unitID.."target"
		if not UnitIsUnit_("target", targetID) then
			targetIndex = GetRaidTargetIndex_(targetID)
			if targetIndex and targetIndex == raidNum then
				FoundPlateGUID(frame, UnitGUID_(targetID), targetID)
				return
			end
		end
	end
end

local function UpdateNameplateInfo(frame)
	if not plateGUIDs[frame] then
		if lib:IsMouseover(frame) then
			GetMouseoverGUID(frame)
		else
			local raidNum = lib:GetRaidIcon(frame)
			if raidNum and raidNum > 0 then
				FindGUIDByRaidIcon(frame, raidNum)
			end
		end
	end

--~ 	debugPrint("Updated nameplate", lib:GetName(frame))
	frame.lnpLastUpdate = 0
end

local function ourOnShow(self, ...)
	lib:SetupNameplate(self)
end

local function ourOnHide(self, ...)
	isOnScreen[self] = false
	RecycleNameplate(self)
end


local function ourOnUpdate(self, elapsed, ...)
	if self.lnpCheckForTarget == true then --Check on the first OnUpdate after the frame's shown.
		self.lnpCheckForTarget = false
		if not plateGUIDs[self] and self:IsShown() and ((self:GetAlpha() == 1) and UnitExists_("target")) then
			FoundPlateGUID(self, UnitGUID_("target"), "target")
		end
	end
	
	self.lnpLastUpdate = (self.lnpLastUpdate or 0) + elapsed
	if self.lnpLastUpdate > updateDelay then
		UpdateNameplateInfo(self)
	end
end


function lib:SetupNameplate(frame)
	isOnScreen[frame] = true
	lib.nameplates[frame] = lib.nameplates[frame] or {}

	if frame:HasScript("OnHide") and not onHideHooks[frame] then
		onHideHooks[frame] = true
		frame:HookScript("OnHide", ourOnHide)
	end
	
	if frame:HasScript("OnShow") and not onShowHooks[frame] then
		onShowHooks[frame] = true
		frame:HookScript("OnShow", ourOnShow)
	end

	if frame:HasScript("OnUpdate") and not onUpdateHooks[frame] then
		onUpdateHooks[frame] = true
		frame:HookScript("OnUpdate", ourOnUpdate)
	end

	if frame:GetNumChildren() == 0 and tidyPlatesRunning == false then
		debugPrint("SetupNameplate 1", "Where'd all the children go?", frame.extended, frame.extended.bars)
	end
	
	if frame:GetNumRegions() == 0 and tidyPlatesRunning == false then
		debugPrint("SetupNameplate 1", "Where'd all the regions go?", frame.extended, frame.extended.regions)
	end

	local LibNP = lib.nameplates[frame]
	local plateName = self:GetName(frame)

	LibNP.name = plateName
	lib.names[plateName] = frame
	
	lib.callbacks:Fire("LibNameplate_NewNameplate", frame)
--~ 	--~ 		debugPrint("LibNameplate_NewNameplate", LibNP:GetName())
	frame.lnpCheckForTarget = true
	UpdateNameplateInfo(frame)
end--	/lnbug




local function CheckForTargetGUID()
	local unitID = "target"
	local GUID
	for frame, LibNP in pairs_(lib.nameplates) do 
		if lib:IsTarget(frame) then
			lib.targeted = frame
			if not plateGUIDs[frame] then
				FoundPlateGUID(frame, UnitGUID_(unitID), unitID)
			end
			lib.callbacks:Fire("LibNameplate_TargetNameplate", frame)
			return
		end
	end
--~ 	debugPrint("Failed to find target GUID")
end



local function OnUpdate(self, elapsed)
	if self.checkTarget == true then
		self.checkTarget = false
		CheckForTargetGUID()
	end

	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate >= scanDelay then
		self.lastUpdate = 0
		FindNameplates()
	end
end



local function OnEvent(self, event, ...)
	if event == "UPDATE_MOUSEOVER_UNIT" then
		if GetMouseFocus_():GetName() == "WorldFrame" then
			for frame in pairs_(lib.nameplates) do 
				
				if frame:IsShown() and lib:IsMouseover(frame) then
					if not plateGUIDs[frame] then
						GetMouseoverGUID(frame)
					end
					lib.callbacks:Fire("LibNameplate_MouseoverNameplate", frame)
					return
				end
			end
		end
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		if UnitExists_("target") then
			self.checkTarget = true--Check for target nameplate on next OnUpdate. We can't do it now.
		else
			lib.targeted = nil
		end
		
	elseif event == "UNIT_TARGET" then
		local unitID = ...
		local targetID = unitID.."target"
		if not UnitIsUnit_("player", unitID) then
			local iconNum = GetRaidTargetIndex_(targetID)
			if iconNum and iconNum > 0 then
				local foundPlate = FindPlateWithRaidIcon(iconNum, targetID)
				local LibNP = lib.nameplates[foundPlate]
				if foundPlate and not plateGUIDs[foundPlate] then
					FoundPlateGUID(foundPlate, UnitGUID_(targetID), targetID)
				end
			end
		end
		
	elseif event == "RAID_TARGET_UPDATE" then
		for frame, LibNP in pairs_(lib.nameplates) do 
			if frame:IsShown() and not plateGUIDs[frame] and lib:IsMarked(frame) then
				local raidNum = lib:GetRaidIcon(frame)
				if raidNum and raidNum > 0 then
					FindGUIDByRaidIcon(frame, raidNum)
				end
			end
		end
	elseif event == "PLAYER_ENTERING_WORLD" then
--~ 		debugPrint(event, MAJOR, _G.TidyPlates, _G.Aloft)
		if _G.TidyPlates then
			tidyPlatesRunning = true
		elseif _G.Aloft then
			aloftRunning = true
		end
		
	end
end


if not lib.frame then
	lib.frame=CreateFrame_("Frame")
end

lib.frame.lastUpdate = 0
lib.frame.checkTarget = false
lib.frame:SetScript("OnUpdate", OnUpdate)
lib.frame:SetScript("OnEvent", OnEvent)
lib.frame:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
lib.frame:RegisterEvent('PLAYER_TARGET_CHANGED')
lib.frame:RegisterEvent('UNIT_TARGET')
lib.frame:RegisterEvent('RAID_TARGET_UPDATE')
lib.frame:RegisterEvent('PLAYER_ENTERING_WORLD')